/' (c) 2022 cepstrum.co.jp           '/
/' This is not free software.        '/
'
' calc fequency response, fs=8kHz
'
' FreeBASIC GUI program
' >fbc -s gui
'

#inclib "openblas"
#inclib "fsconv"

#include once "portaudio.bi"
#include once "fbblas.bi"
#include once "fsconv.bi"
#include once "graphics.bi"
#include once "crt.bi"
#include once "string.bi"
#include once "windows.bi"

#define TSPLEN 2048
#define BUFLEN 2048
#define FFTLEN 2048

#include once "fft.bi"

dim shared as single ar(TSPLEN)
dim shared as single ai(TSPLEN)
dim shared as single tsp(TSPLEN)

#define FS          48000
#define PA_BUFLEN   6         ' 6=48kHz/8kHz

#define NOISE_LEVEL 0.2

dim shared as single  xcopy(BUFLEN)
dim shared as integer xcopy_wt_busy = 0
dim shared as integer xcopy_rd_busy = 1
dim shared as integer cmd_opt_val_t = -1

type User_Data_Type
  frame_count as integer
end type

dim user_data        as User_Data_Type
dim outputParameters as PaStreamParameters 
dim inputParameters  as PaStreamParameters 
dim stream           as any ptr 
dim errcode          as integer<32>

dim now_time         as string
dim past_time        as string
dim past_frame_count as integer

dim asio_api_index   as integer
dim host_api_count   as integer
dim host_api_info    as PaHostApiInfo ptr
dim device_index     as integer
dim device_name      as zstring * 512

dim inchar           as string
dim keycode          as integer

dim plotx            as integer
dim ploty            as integer

dim i                as integer

dim cmd_opt_val_d    as integer
dim cmd_opt_val_b    as integer

dim latency          as double

dim c as single

'==========================================================================

function delay2z(x as single) as single
  dim    y   as single
  static zzx as single = 0.0
  static zx  as single = 0.0

  y  =zzx
  zzx=zx
  zx =x
  return y
end function

'==========================================================================

function bpf2nd20hz2nd23980hz(x as single) as single
  dim    y     as single
  static zzzzx as single = 0.0
  static zzzx  as single = 0.0
  static zzx   as single = 0.0
  static zx    as single = 0.0
  static zzzzy as single = 0.0
  static zzzy  as single = 0.0
  static zzy   as single = 0.0
  static zy    as single = 0.0

  y=9.963044429897593979D-01*(x-2.0*zzx+zzzzx)+1.992595228843776267D+00*zzy-9.926225431269650734D-01*zzzzy
  zzzzx=zzzx
  zzzx =zzx
  zzx  =zx
  zx   =x
  zzzzy=zzzy
  zzzy =zzy
  zzy  =zy
  zy   =y
  return y
end function

'==========================================================================

function hpf4th250hz(x as single) as single
  static zzzzx as single = 0.0
  static zzzx  as single = 0.0
  static zzx   as single = 0.0
  static zx    as single = 0.0
  static zzzzy as single = 0.0
  static zzzy  as single = 0.0
  static zzy   as single = 0.0
  static zy    as single = 0.0
  dim    y     as single

  y=7.733467891606214950e-01*(x-4.0*zx+6.0*zzx-4.0*zzzx+zzzzx)+3.487307741549900086e+00*zy-4.589291232078407390e+00*zzy+2.698884391340754973e+00*zzzy-5.980652616008872435e-01*zzzzy
  zzzzx=zzzx
  zzzx =zzx
  zzx  =zx
  zx   =x
  zzzzy=zzzy
  zzzy =zzy
  zzy  =zy
  zy   =y
  return y
end function

'==========================================================================

function lpf2nd3960hz(x as single) as single
  static zzx as single = 0.0
  static zx  as single = 0.0
  static zzy as single = 0.0
  static zy  as single = 0.0
  dim    y   as single

  y=9.780304792065596330E-1*(x+2.0*zx+zzx)-1.955578240315035470E0*zy-9.565436765112032450E-1*zzy
  zzx=zx
  zx =x
  zzy=zy
  zy =y
  return y
end function

'==========================================================================

function callbackfunc cdecl(                   _
  pread      as single ptr,                    _
  pwrite     as single ptr,                    _
  rwcount    as uinteger<32>,                  _
  timeInfo   as PaStreamCallbackTimeInfo ptr,  _
  flag       as uinteger<32>,                  _
  user_data  as User_Data_Type ptr             _
) as integer

  static frame_counter as integer = 0

  dim ad48k(6) as LrData
  dim da48k(6) as LrData
  dim ad16k0   as LrData
  dim ad16k1   as LrData
  dim da16k0   as LrData
  dim da16k1   as LrData
  dim ad8k     as LrData
  dim da8k     as LrDAta
  dim x        as single
  dim i        as integer

  static xbuf(BUFLEN) as single
  static init_flag    as integer = 1
  static tsp_ix       as integer = 0
  static xbuf_ix      as integer = 0

  if init_flag=1 then
    init_flag=0
    for i=0 to BUFLEN-1
      xbuf(i)=0.0
    next
    return 0
  end if

  '-------------------------------------------------------------------------------------
  ' read A/D and Fs conversion (48kHz ---> 8kHz)

  for i=0 to 5
    ad48k(i).l=*(pread+2*i+0)
    ad48k(i).r=*(pread+2*i+1)
  next
  decimate6to2(ad48k(0), ad48k(1), ad48k(2), ad48k(3), ad48k(4), ad48k(5), ad16k0, ad16k1)    ' 6:2 decimation, fs=48kHz --> 16kHz
  decimate2iir_lpc(ad16k0, ad16k1, ad8k)       

  '-------------------------------------------------------------------------------------

  xbuf_ix=(xbuf_ix+1) mod TSPLEN

  x=ad8k.l
  x=hpf4th250hz(x)
  x=lpf2nd3960hz(x)
  xbuf(xbuf_ix)=x

  if cmd_opt_val_t=0 then
    xbuf(xbuf_ix)=bpf2nd20hz2nd23980hz(delay2z(tsp(xbuf_ix)))          ' internal (software) loopback test
  end if

  tsp_ix=(tsp_ix+1) mod TSPLEN
  da8k.l=NOISE_LEVEL*tsp(tsp_ix)
  da8k.r=0.0

  '-------------------------------------------------------------------------------------
  ' Fs conversion (8kHz ---> 48kHz) amd wrote D/A

  interpolate2iir_lpc(da8k, da16k0, da16k1)                                                    ' 1:2 interpolation, fs=8kHz --> 16kHz
  interpolate2to6(da16k0, da16k1, da48k(0), da48k(1), da48k(2), da48k(3), da48k(4), da48k(5))  ' 2:6 interpolation, fs=16kHz --> 48kHz
  for i=0 to 5
    *(pwrite+2*i+0)=da48k(i).l
    *(pwrite+2*i+1)=da48k(i).r
  next

  '-------------------------------------------------------------------------------------

  if xcopy_rd_busy=0 then
    xcopy_wt_busy=1
    for i=0 to BUFLEN-1
      xcopy(i)=xbuf(i)
    next
    xcopy_wt_busy=0
  end if

  frame_counter=frame_counter+1
  (user_data->frame_count)=frame_counter

  return 0

end function

'==========================================================================

function cmd_opt_int(optstr as string) as integer
  dim s       as string
  dim i       as integer
  dim opt_val as integer
  
  opt_val=-1
  i=1
  while command(i)<>""
    s=lcase(command(i)) 
    if instr(s, "/"+optstr)<>0 or instr(s, "-"+optstr) then
       s=mid(s, 3)
       opt_val=valuint(s)
    end if  
    i=i+1
  wend
  return opt_val
end function

'=====================================================================

sub device2name(device_index as integer, byref strbuf as zstring ptr) 
  dim device_info as PaDeviceInfo ptr
  dim wstrbuf     as wstring * 512

  device_info=Pa_GetDeviceInfo(device_index)
  MultiByteToWideChar(CP_UTF8, 0, device_info->name, -1, wstrbuf, 512)     ' --> unicode
  WideChartoMultiByte(CP_ACP, 0, @wstrbuf, 512, strbuf, 512, NULL, NULL)   ' --> shiftjis
end sub

'==========================================================================

sub plot_thread(ar() as single, opt_b as integer, device_str as zstring)
  static zploty as integer
  dim plotx as integer
  dim ploty as integer
  dim i     as integer
  dim aravr as single
  dim s     as string
  
  g_fillbox(0, 0, 1024+1, 600+1, BLACK_COLOR)       ' clear screen
  g_line(0, 100, 1024, 100, DARK_GRAY_COLOR)
  g_line(0, 200, 1024, 200, DARK_GRAY_COLOR)
  g_line(0, 300, 1024, 300, DARK_GRAY_COLOR)
  g_line(0, 400, 1024, 400, DARK_GRAY_COLOR)
  g_line(0, 500, 1024, 500, DARK_GRAY_COLOR)

  g_dotline(128, 0, 128, 600, DARK_GRAY_COLOR, &Haaaaaaaa) 
  g_line   (256, 0, 256, 600, DARK_GRAY_COLOR)
  g_dotline(384, 0, 384, 600, DARK_GRAY_COLOR, &Haaaaaaaa) 
  g_line   (512, 0, 512, 600, DARK_GRAY_COLOR)
  g_dotline(640, 0, 640, 600, DARK_GRAY_COLOR, &Haaaaaaaa) 
  g_line   (768, 0, 768, 600, DARK_GRAY_COLOR)
  g_dotline(896, 0, 896, 600, DARK_GRAY_COLOR, &Haaaaaaaa) 

  aravr=0.0
  for i=100 to FFTLEN/2-1-100
    aravr=aravr+(ar(i))
  next
  aravr=aravr/(FFTLEN/2.0-200.0)

  for i=0 to FFTLEN/2-1
    plotx=i
    'ploty=(ar(i)-aravr)*100+200     ' 1dB/div.
    ploty=(ar(i)-aravr)*10+300     ' 10dB/div.
    if i=0 then
      zploty=ploty
    end if
    g_line(plotx-1, zploty, plotx, ploty, GREEN_HIGH_COLOR)
    zploty=ploty
  next

  g_string_bg(  2, 2, "0",    BLACK_COLOR) 
  g_string   (  2, 2, "0",    LIGHT_GRAY_COLOR) 
  g_string_bg(241, 2, "1kHz", BLACK_COLOR) 
  g_string   (241, 2, "1kHz", LIGHT_GRAY_COLOR) 
  g_string_bg(498, 2, "2kHz", BLACK_COLOR) 
  g_string   (498, 2, "2kHz", LIGHT_GRAY_COLOR) 
  g_string_bg(754, 2, "3kHz", BLACK_COLOR) 
  g_string   (754, 2, "3kHz", LIGHT_GRAY_COLOR) 
  g_string_bg(992, 2, "4kHz", BLACK_COLOR) 
  g_string   (992, 2, "4kHz", LIGHT_GRAY_COLOR) 

  g_tofu  (990, 590, 4, BLACK_COLOR)
  g_string(990, 590, "30dB", LIGHT_GRAY_COLOR) 
  g_tofu  (990, 496, 4, BLACK_COLOR)
  g_string(990, 496, "20dB", LIGHT_GRAY_COLOR) 
  g_tofu  (990, 396, 4, BLACK_COLOR)
  g_string(990, 396, "10dB", LIGHT_GRAY_COLOR) 
  g_tofu  (998, 296, 3, BLACK_COLOR)
  g_string(998, 296, "0dB", LIGHT_GRAY_COLOR) 
  g_tofu  (982, 196, 5, BLACK_COLOR)
  g_string(982, 196, "-10dB", LIGHT_GRAY_COLOR) 
  g_tofu  (982,  96, 5, BLACK_COLOR)
  g_string(982,  96, "-20dB", LIGHT_GRAY_COLOR) 

  g_string_bg(2, 42, "hit ESC or ctrl-c to exit", BLACK_COLOR) 
  g_string   (2, 42, "hit ESC or ctrl-c to exit", ORANGE_COLOR) 

  s=device_str
  g_tofu(2, 26, len(s), BLACK_COLOR)
  g_string(2, 26, s, YELLOW_COLOR) 

  s="buffer size : "+format(opt_b, "0")
  g_tofu(2, 16, len(s), BLACK_COLOR)
  g_string(2, 16, s, YELLOW_COLOR) 

end sub

'==========================================================================

windowtitle "ASIO frequency response fs=8kHz, (c) 2022 cepstrum.co.jp"
screenres 1024+1, 600+1

if (cmd_opt_int("h")<>-1) or (cmd_opt_int("?")<>-1) then          ' if /h, -h, /H, -H, /?, -? option
  print " ASIO realtime frequency response measurement program, (c) 2022 cepstrum.co.jp"
  print
  print "option"
  print "  -d : set device index of PortAudio"
  print "  -b : set buffer size (8/16/32/64/128/256/512/1024/2048)"
  print "  -t : internal (software) loopback test"
  print 
  print "example"
  print "  >asio_rt_response30db -d3 -b128"
  print 
  print "hit any key to exit"
  sleep
  system
end if

cmd_opt_val_d=cmd_opt_int("d")          ' read /d, -d, /D, -D option

cmd_opt_val_b=cmd_opt_int("b")          ' read /b, -b, /B, -B option
select case cmd_opt_val_b
  case 8
    latency=0.000166
  case 16
    latency=0.000333
  case 32
    latency=0.000666
  case 64
    latency=0.001333
  case 128
    latency=0.002666
  case 256
    latency=0.005333
  case 512
    latency=0.010066
  case 1024
    latency=0.021333
  case 2048
    latency=0.042666
  case else
    latency=0.001333
    cmd_opt_val_b=64
end select

cmd_opt_val_t=cmd_opt_int("t")          ' read /t, -t, /T, -T option

'--------------------------------------------------------------------------
' generate TSP signal

fft(ar(), ai(), -1, 1)        ' initialize FFT routine

ar(TSPLEN\2)=1.0
ai(TSPLEN\2)=0.0
for i=0 to TSPLEN\2-1
  c=-2.0*3.1415926*cdbl(i)*cdbl(i)/cdbl(TSPLEN)
  ar(i)=cos(c)
  ai(i)=sin(c)
next            
for i=1 to TSPLEN\2-1
  ar(TSPLEN-i)=ar(i)
  ai(TSPLEN-i)=-ai(i)
next
fft(ar(), ai(), 1, 0)   ' generate TSP by IFFT

for i=0 to TSPLEN-1
  tsp(i)=0.95*ar(i)*sqrt(TSPLEN/2.0)      ' scaling (-1.0<tsp(i)<1.0)
next

'--------------------------------------------------------------------------

errcode=Pa_Initialize()
if errcode<>0 then
  Pa_Terminate()
  print "!!! PortAudio initialize error"
  sleep 3000
  system
end if

if cmd_opt_val_d<0 then
  asio_api_index=Pa_HostApiTypeIdToHostApiIndex(3)     ' (index val 3)= asio
  host_api_info=Pa_GetHostApiInfo(asio_api_index)
  if (host_api_info->deviceCount)=0 then
    Pa_Terminate()
    print "!!! no ASIO device"
    sleep 3000
    system   
  end if
  device_index=Pa_HostApiDeviceIndexToDeviceIndex(asio_api_index, 0)   ' find 1st ASIO device
else
  device_index=cmd_opt_val_d
end if

device2name(device_index, @device_name)

outputParameters.device                   =device_index
outputParameters.channelCount             =2
outputParameters.sampleFormat             =paFloat32 
outputParameters.suggestedLatency         =paLatencyFs48k64Samples
outputParameters.hostApiSpecificStreamInfo=NULL

inputParameters.device                    =device_index
inputParameters.channelCount              =2
inputParameters.sampleFormat              =paFloat32
inputParameters.suggestedLatency          =paLatencyFs48k64Samples
inputParameters.hostApiSpecificStreamInfo =NULL

errcode=Pa_OpenStream(  _  
  @stream,              _  ' PaStream  ptr ptr
  @inputParameters,     _  ' input parameters
  @outputParameters,    _  ' output parameters
  FS,                   _  ' Fs
  PA_BUFLEN,            _  ' frames per buffer
  0,                    _  ' stream flags
  @callbackfunc,        _  ' callback function
  @user_data            _  ' user data
) 
if errcode<>0 then
  Pa_Terminate()
  print "!!! PortAudio stream setting error"
  sleep 3000
  system
end if

errcode=Pa_StartStream(stream)
if errcode<>0 then
  Pa_Terminate()
  print "!!! PortAudio can't start stream"
  sleep 3000
  system
end if

past_time=""
past_frame_count=0
do

  if xcopy_wt_busy=0 then
    xcopy_rd_busy=1

    for i=0 to FFTLEN-1
      ar(i)=xcopy(i)
      ai(i)=0.0
    next

    fft(ar(), ai(), -1, 0)

    for i=0 to FFTLEN/2-1
      ar(i)=ar(i)*ar(i)+ai(i)*ai(i)   ' calc power
      ar(i)=10.0*log10(ar(i)+1e-19)   ' cals power spectrum
    next

    dim thread_id as any ptr = threadcall plot_thread(ar(), cmd_opt_val_b, device_name)

    sleep 800, 1
    xcopy_rd_busy=0
  end if

  sleep 200, 1
  inchar=inkey$
 
  keycode=asc(inchar)
  if keycode=3 or keycode=27 then           ' if ctrl-C or ESC
    errcode=Pa_StopStream(stream)
    errcode=Pa_CloseStream(stream)
    Pa_Terminate()
    system
  end if

  now_time=time
  if now_time<>past_time then
    past_frame_count=user_data.frame_count
  end if
  past_time=now_time
loop


